//////////////////////////////////////////////////////////////////////////////////////
// File     : x16demo.c																//
// Function : Linux demo application for Sensoray model 516/616 boards				//
// Author   : Jim Lamberson and Charlie Liu											//
// Revision :																		//
//            Sept, 2003   Jim Lamberson    release sx16-0.7						//
//            Oct., 2003   Charlie X. Liu   Added demo for using caption overlay	//
//																					//
//////////////////////////////////////////////////////////////////////////////////////

/*
Copyright (C) 2003  Sensoray Co., Inc.  This program is free software;
you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

You must obey the GNU General Public License in all respects for all of
the code used. If you modify this file, you may extend this exception
to your version of the file, but you are not obligated to do so. If you
do not wish to do so, delete this exception statement from your version.
*/

// The necessary headers /////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <time.h>
#include <sys/ioctl.h>

#include "sx16ioctl.h"			    // ioctl definitions for sx16 driver.
#include "sx16core.h"               // x16 attributes.


// Manifest constants ////////////////////////////////////////////////////////////////////////////////

#define ENABLED         1
#define DISABLED        0

#define TRUE            1
#define FALSE           0

#define MAXLEN          80              // Maximum length of user command string.
#define MAXNUMBOARDS    4               // Maximum number of boards supported by the driver.

// Stream mode:
#define MODE_IDLE       0               // Halted.
#define MODE_PLAY       1               // Streaming from file.
#define MODE_RECORD     2               // Streaming to file.


// Configuration constants ///////////////////////////////////////////////////////////////////////////

#define MAX_BOARDS		4 				// Maximum number of x16 boards in the system.
#define MAX_DEVS		16 				// Maximum number of v4l devices in the system.
#define MPEG_BUF_SIZE	4096			// Number of MPEG bytes to transfer in a single chunk.  Max size is 32K.
#define BLOCKING_MODE	TRUE			// Set to TRUE to use blocking read/write operations.

// Device name for the x16 boards: v4l.
#define DEV_NAME		"/dev/sx16a"


// Storage //////////////////////////////////////////////////////////////////

typedef struct BOARD {					// BOARD OBJECT -------
  int				handle;		        //  Handle to x16 device.
  int				mode;		        //  Stream mode.
  int				stop_rq;	        //  Stream Halt requested.
  char 				devname[MAXLEN];	//  Device name for board.
  char 				fname[MAXLEN];	    //  File name for mpeg data.
  pthread_t			io_thread_id;		//  Handle to stream i/o thread.
  unsigned int	mpeg_count;			    //  Number of MPEG bytes transferred.
  int						flags;					//  Status flags from i/o thread to main thread.
} BOARD;

static struct BOARD bd[MAX_BOARDS];		// Board objects.
static int num_boards;                  // Actual number of boards.
static int act_bdnum	= 0;            // Board 0 is active by default.


// Function prototypes ///////////////////////////////////////////////////////////////////////////////

static int app_init( void );
static void app_shutdown( void );
static void cmd_processor( void );
static void *io_thread_func( void* arg );
static void set_vsystem( struct BOARD *b, int sys );
static void get_attributes( struct BOARD *b );
static void set_attribute( struct BOARD *b, int val, int minval, int maxval, int ioctl_num );
static void get_caption_settings( struct BOARD *b );
static void start_stream( struct BOARD *b, char *fname, int new_stream_mode );
static void stop_stream( struct BOARD *b );
static void int_test( struct BOARD *b, int num );


///////////////////////////////////////////////////////////////////////////
// Init the system and execute user commands until termination.

int main( int argc, char *argv[] )
{
  int rtnval;

  printf("starting");
  // Attempt to open and initialize all boards.
  rtnval = app_init();

  // If any board successfully opened and initialized ...
  if ( rtnval > 0 )
  {
    // Execute user commands until Application Quit command is invoked.
    cmd_processor();
  }

  // Close all opened boards.
  app_shutdown();

  // Exit application.
  return 0;
}

///////////////////////////////////////////////////////////////////////
// Ceck board type.  Returns: 516/616 for sx16, 0 in no board or any other code.

static int check_type( char *dev_name )
{
  int handle,		        // Handle to a device.
      type;                 // Type of board.

  // Open device.
  handle = open( dev_name, O_RDWR );
  if ( handle == -1 )
  {
    return( 0 );
  }
  type = ioctl( handle, SX16_IOC_GET_BTYPE, 0 );
  close( handle );

  return type;
}

///////////////////////////////////////////////////////////////////////
// Open and initialize one board.  Returns: 0=success, -1=error.

static int board_open( int board_num, char *dev_name )
{
  // Set pointer to board's object.
  struct BOARD *b = &bd[board_num];

  // Initialize object.
  b->mode			= MODE_IDLE;
  b->fname[0]	= 0;
  b->flags		= 0;

  // Open device.
  b->handle = open( dev_name, O_RDWR | ( BLOCKING_MODE ? 0 : O_NONBLOCK ) );
  if ( b->handle == -1 )
  {
    printf( "Driver open error\n" );
    return( -1 );
  }

  strcpy( b->devname, dev_name );
  // Program the default settings.
  ioctl( b->handle, SX16_IOC_SET_COLORBARS,			DISABLED	);
  ioctl( b->handle, SX16_IOC_SET_VIDEO_SYSTEM,	    SYS_NTSC	);
  ioctl( b->handle, SX16_IOC_SET_BITRATE,			1000000		);
  ioctl( b->handle, SX16_IOC_SET_MPEG,				3	);
  ioctl( b->handle, SX16_IOC_SET_INPUT,             0	);
  ioctl( b->handle, SX16_IOC_SET_INPUT_TYPE,		CVIDEO		);
  ioctl( b->handle, SX16_IOC_SET_VBR,				DISABLED	);
  ioctl( b->handle, SX16_IOC_SET_AUDIO_VOLUME,	    6			);
  ioctl( b->handle, SX16_IOC_SET_AUDIO_MUTE,		DISABLED	);

  // Set the caption defaults.
  ioctl( b->handle, SX16_IOC_SET_CAPTION_WINDOW,			1				);	// window #1
  ioctl( b->handle, SX16_IOC_SET_CAPTION_XSTART,			100				);	// 100
  ioctl( b->handle, SX16_IOC_SET_CAPTION_YSTART,			10				);	// 10
  ioctl( b->handle, SX16_IOC_SET_CAPTION_SCALE,				OVL_SCALE_X1	);	// x1
  ioctl( b->handle, SX16_IOC_SET_CAPTION_COLOR,				OVL_WHITE		);	// white
  ioctl( b->handle, SX16_IOC_SET_CAPTION_TRANSPARENCY,		OVL_TRANSP_T	);	// transparent
  ioctl( b->handle, SX16_IOC_SET_CAPTION_ON,				CAPTION_OFF		);	// off
  ioctl( b->handle, SX16_IOC_SET_CAPTION_TEXT,              "SX16 Demo"     );  // text
  return 0;
}

///////////////////////////////////////////////////////////////////////
// Open and initialize all boards.  Returns: number of boards.

static int app_init( void )
{
  int		i;
  char	dev_name[80];
  int type;

  // For each board ...
  for ( i = num_boards = 0; i < MAX_DEVS; i++ )
  {
    // Specify board's device name.
    sprintf( dev_name, DEV_NAME "%d", i );
    // Check board type.  Go to next if not sx16.
    type = check_type( dev_name );
    if ( type != 516 && type != 616 )
    {
        if( type != 0 )
            printf( "Not a sx16 board #%d found as %s\n", num_boards, dev_name );
        continue;
    }
    // Open the board. Go to next if fail.
    if( board_open( num_boards, dev_name ) )
      continue;
    printf( "%d board #%d found as %s\n", type, num_boards, dev_name );
    if( num_boards == MAX_BOARDS - 1 )
      break;
    num_boards++;
  }

  return num_boards;
}

//////////////////////////////////////////////////////////////
// Terminate all stream threads and close all boards.

static void app_shutdown( void )
{
  struct BOARD *b;

  // For each board ...
  for ( b = bd; b < &bd[num_boards]; b++ )
  {
    // We are finished if this board was never opened.
    if ( b->handle == -1 )
      break;

    // Terminate the board's I/O thread if it exists.
    stop_stream( b );

    // Close the board.
    close( b->handle );
  }

  printf( "\n" );
}

//////////////////////////////////////////////////////////////
// Display the user prompt.

static void prompt( void )
{
  printf( "x16demo[%d]>", act_bdnum );
  fflush( stdout );
}

//////////////////////////////////////////////////////////////
// Display the command menu.

static void show_menu( void )
{
  printf( "+-SX16DEMO-COMMANDS------------------------------------------------------------+\n" );
  printf( "|   board n ....... select board: 0..n                                         |\n" );
  printf( "|   chan n ........ set video channel: 0 to 3                                  |\n" );
  printf( "|   get ........... display current settings                                   |\n" );
  printf( "|   intype n ...... set video type: 0=S-Video, 1=Composite                     |\n" );
  printf( "|   loopback ...... set video codec loopback (516 only): 0|1                   |\n" );
  printf( "|   rgpio n ....... display GPIO output states                                 |\n" );
  printf( "|   wgpio n ....... set GPIO output states: 0=off, 1=on                        |\n" );
  printf( "|   h[elp] ........ display this list                                          |\n" );
  printf( "|   q[uit] ........ quit program                                               |\n" );
  printf( "|   reset ......... reset board                                                |\n" );
  printf( "+-AUDIO------------------------------------------------------------------------+\n" );
  printf( "|   arate n ....... set bits/second:                                           |\n" );
  printf( "|                     1=32K, 2=48K, 3=56K, 4=64K, 5=80K, 6=96K, 7=112K, 8=128K |\n" );
  printf( "|                     9=160K, 10=192K, 11=224K, 12=256K, 13=320K, 14=384K      |\n" );
  printf( "|   aformat n ..... set format:                                                |\n" );
  printf( "|                     0=mpeg1/layer1, 1=mpeg1/layer2                           |\n" );
  printf( "|   mute n ........ set mute enable: 0=no mute, 1=mute                         |\n" );
  printf( "|   vol n ......... set volume: 0..63 (-dB units)                              |\n" );
  printf( "+-VIDEO------------------------------------------------------------------------+\n" );
  printf( "|   bars n ........ set output colorbars enable: 0=off, 1=on                   |\n" );
  printf( "|   bright n ...... set luminance brightness: 0..255                           |\n" );
  printf( "|   contrast n .... set luminance contrast: 0..255                             |\n" );
  printf( "|   vsat n ........ set chrominance saturation: 0..255                         |\n" );
  printf( "|   vsystem n ..... set video system: 1=PAL, 2=NTSC                            |\n" );
  printf( "|   vstatus ....... query video status                                         |\n" );
  printf( "+-MPEG-------------------------------------------------------------------------+\n" );
  printf( "|   gop n ......... set GOP: 0 to 31                                           |\n" );
  printf( "|   ipdist n ...... set IP distance: 0 to 3                                    |\n" );
  printf( "|   mpeg n ........ set mpeg type and picture size:                            |\n" );
  printf( "|                     0=mpeg2:D1-704, 1=mpeg2:2/3 D1, 2=mpeg2:1/2 D1,          |\n" );
  printf( "|                     3=mpeg1:SIF,    4=mpeg1:QSIF,                            |\n" );
  printf( "|                     5=mpeg2:D1-720, 6=mpeg2:D1-640                           |\n" );
  printf( "|   stream n ...... set stream type: 0=mux, 1=video es, 2=audio es             |\n" );
  printf( "|   avsel n ....... set mux a/v select: 0=aud+vid, 1=vid only, 2=aud only      |\n" );
  printf( "|   vrate n ....... set video bits/second: 100000..10000000                    |\n" );
  printf( "|   vbr n ......... set variable bit rate enable: 0=off, 1=on                  |\n" );
  printf( "+-STREAM-----------------------------------------------------------------------+\n" );
  printf( "|   play fname .... play from MPEG file; default name is \"tmp.mpg\"             |\n" );
  printf( "|   rec fname ..... record to MPEG file; default name is \"tmp.mpg\"             |\n" );
  printf( "|   stop .......... stop MPEG stream and close file                            |\n" );
  printf( "+-CAPTION----------------------------------------------------------------------+\n" );
  printf( "|   window n ...... set window number: 1 or 2                                  |\n" );
  printf( "|   xstart n ...... set window position for x_start: 8..700                    |\n" );
  printf( "|   ystart n ...... set window position for y_start: 0..478 / 574 (NTSC/PAL)   |\n" );
  printf( "|   scale n ....... set scale for font:                                        |\n" );
  printf( "|                     0=x1, 1=x2, 2=x2(only vertical), 3=x4                    |\n" );
  printf( "|   color n ....... set color for caption:                                     |\n" );
  printf( "|                     0=white, 1=black,                                        |\n" );
  printf( "|   transp n ...... set transparency attribute for caption:                    |\n" );
  printf( "|                     1=transparent,     2=black-background                    |\n" );
  printf( "|                     3=transp-text on white-bkgrnd                            |\n" );
  printf( "|                     4=black-text on white-bkgrnd                             |\n" );
  printf( "|   text string ... set text string for overlay:  any ascii string             |\n" );
  printf( "|   caption n ..... set caption off/on: 0=off, 1=on                            |\n" );
  printf( "+------------------------------------------------------------------------------+\n" );
  printf( "+-SPECIAL TESTS----------------------------------------------------------------+\n" );
  printf( "|   int n ......... wait for n video interrupts                                |\n" );
  printf( "+------------------------------------------------------------------------------+\n" );
}

///////////////////////////////////////////////////////////////////////////////
// Decode and execute user command.  Returns TRUE if Quit selected.

static int exec_command( char *cmd, char *arg )
{
  struct BOARD	*b;

  // Convert argument to integer value (not relevant for all commands).
  int arg_val = -1;
  sscanf( arg, "%d", &arg_val );

  // Decode and execute command -----------------------------------

  // Terminate application.
  if ( cmd[0] == 'q' )
    return TRUE;

  // Select board.
  if ( strcmp( cmd, "board" ) == 0 )
  {
    if ( ( arg_val >= 0 ) && ( arg_val < num_boards ) )
      act_bdnum = arg_val;
    else
      printf( "Illegal or missing argument; must be in range 0..%d\n", num_boards - 1 );
    prompt();
    return FALSE;
  }

  // Show help.
  if ( ( cmd[0] == 'h' ) || ( cmd[0] == '?' ) )
  {
    show_menu();
    prompt();
    return FALSE;
  }

  // Prompt for user command.
  prompt();

  // Set pointer to active board's object.
  b = &bd[act_bdnum];

  // Set video attribute.
  if      ( strcmp( cmd, "bright"    ) == 0 )	set_attribute( b, arg_val, 0,	                255,        				SX16_IOC_SET_BRIGHTNESS    );
  else if ( strcmp( cmd, "intype"    ) == 0 )	set_attribute( b, arg_val, CVIDEO,	         	SVIDEO,    					SX16_IOC_SET_INPUT_TYPE    );
  else if ( strcmp( cmd, "chan"      ) == 0 )	set_attribute( b, arg_val, 0,					        3,          				SX16_IOC_SET_INPUT			   );
  else if ( strcmp( cmd, "bars"      ) == 0 )	set_attribute( b, arg_val, DISABLED,	        ENABLED,    				SX16_IOC_SET_COLORBARS		 );
  else if ( strcmp( cmd, "contrast"  ) == 0 )	set_attribute( b, arg_val, 0,	                255,		    				SX16_IOC_SET_CONTRAST      );
  else if ( strcmp( cmd, "vstatus"   ) == 0 )
    {
      int sts = ioctl( b->handle, SX16_IOC_GET_INPUT_STATUS, 0 );
      printf( "sts=%x (%s)\n", sts, (sts & VIDEO_DET) == 0 ? "No video signal" : sts & F60HZ_DET ? "NTSC video detected" : "PAL video detected" );
    }
  else if ( strcmp( cmd, "vsystem"   ) == 0 )	set_vsystem( b, arg_val );

	// Set mpeg codec attribute.
  else if ( strcmp( cmd, "mpeg"      ) == 0 )	set_attribute( b, arg_val, SIZE704_2,	        SIZE640_2,  				SX16_IOC_SET_MPEG		       );
	else if ( strcmp( cmd, "stream"    ) == 0 )	set_attribute( b, arg_val, STREAMTYPE_MUX,    3,			SX16_IOC_SET_STREAM_TYPE   );
  else if ( strcmp( cmd, "vrate"     ) == 0 )	set_attribute( b, arg_val, 100000,		        20000000,   				SX16_IOC_SET_BITRATE		   );
  else if ( strcmp( cmd, "vbr"       ) == 0 )	set_attribute( b, arg_val, DISABLED,	        ENABLED,    				SX16_IOC_SET_VBR			     );
  else if ( strcmp( cmd, "vsat"      ) == 0 )	set_attribute( b, arg_val, 0,	                255,		    				SX16_IOC_SET_SATURATION    );
  else if ( strcmp( cmd, "avsel"     ) == 0 )	set_attribute( b, arg_val, 0,	                2,	  		  				SX16_IOC_SET_AV_SELECT	   );
  else if ( strcmp( cmd, "prognum"   ) == 0 )	set_attribute( b, arg_val, 0,	                255,  		  				SX16_IOC_SET_PROG_NUM	     );  // not in menu
  else if ( strcmp( cmd, "gop"       ) == 0 )	set_attribute( b, arg_val, 0,	                31,	  	    				SX16_IOC_SET_M             );
  else if ( strcmp( cmd, "ipdist"    ) == 0 )	set_attribute( b, arg_val, 0,	                3,	  	    				SX16_IOC_SET_N             );

  // Set audio attribute.
  else if ( strcmp( cmd, "vol"       ) == 0 )	set_attribute( b, arg_val, 0,					63,     		 			SX16_IOC_SET_AUDIO_VOLUME  );
  else if ( strcmp( cmd, "mute"      ) == 0 )	set_attribute( b, arg_val, DISABLED,	        ENABLED,		 			SX16_IOC_SET_AUDIO_MUTE    );
  else if ( strcmp( cmd, "arate"     ) == 0 )	set_attribute( b, arg_val, 1,     		        14,        					SX16_IOC_SET_AUDIO_BITRATE );
  else if ( strcmp( cmd, "aformat"   ) == 0 )	set_attribute( b, arg_val, AFMT_MPEG1_LAYER1, AFMT_MPEG1_LAYER1, 			SX16_IOC_SET_AUDIO_FORMAT  );

  // Set/get GPIO output states.
  else if ( strcmp( cmd, "wgpio"     ) == 0 ) ioctl( b->handle, SX16_IOC_SET_GPIO, arg_val );
  else if ( strcmp( cmd, "rgpio"     ) == 0 ) printf( "gpio=%d\n", ioctl( b->handle, SX16_IOC_GET_GPIO, arg_val ) );

	// Display current settings.
  else if ( strcmp( cmd, "get"       ) == 0 )	{ get_attributes( b ); get_caption_settings( b); prompt(); }

  // Control MPEG stream mode.
  else if ( strcmp( cmd, "rec"       ) == 0 )	start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_RECORD );
  else if ( strcmp( cmd, "play"      ) == 0 )	start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY   );
  else if ( strcmp( cmd, "stop"      ) == 0 )	stop_stream( b );

  // Set Caption attributes/parameters.
  else if ( strcmp( cmd, "window"    ) == 0 )	set_attribute( b, arg_val, 1,				2,					SX16_IOC_SET_CAPTION_WINDOW );
  else if ( strcmp( cmd, "xstart"    ) == 0 )	set_attribute( b, arg_val, 8,				MAX_XSTART,			SX16_IOC_SET_CAPTION_XSTART );
  else if ( strcmp( cmd, "ystart"    ) == 0 )	set_attribute( b, arg_val, 0,				MAX_YSTART_PAL,		SX16_IOC_SET_CAPTION_YSTART );
  else if ( strcmp( cmd, "scale"     ) == 0 )	set_attribute( b, arg_val, OVL_SCALE_X1,	OVL_SCALE_X4,		SX16_IOC_SET_CAPTION_SCALE );
  else if ( strcmp( cmd, "color"     ) == 0 )	set_attribute( b, arg_val, OVL_WHITE,		OVL_BLACK,			SX16_IOC_SET_CAPTION_COLOR );
  else if ( strcmp( cmd, "transp"    ) == 0 )	set_attribute( b, arg_val, OVL_TRANSP_T,	OVL_TRANSP_B_BKGRD,	SX16_IOC_SET_CAPTION_TRANSPARENCY );
  else if ( strcmp( cmd, "text"      ) == 0 )	ioctl( b->handle, SX16_IOC_SET_CAPTION_TEXT, arg );
  else if ( strcmp( cmd, "caption"   ) == 0 )	set_attribute( b, arg_val, CAPTION_OFF,		CAPTION_ON,			SX16_IOC_SET_CAPTION_ON );

	// Other commands.
  else if ( strcmp( cmd, "loopback"  ) == 0 ) ioctl( b->handle, SX16_IOC_GA_LOOPBACK, arg_val );
  else if ( strcmp( cmd, "reset"     ) == 0 ) ioctl( b->handle, SX16_IOC_RESET, 0 );
  else if ( strcmp( cmd, "int"       ) == 0 ) int_test( b, arg_val );
  // Illegal command.
  else if ( cmd[0] != 0 )
  {
    printf( "Invalid command: %s\n", cmd );
    prompt();
  }

  return FALSE;
}

////////////////////////////////////////////////////////////////////
// Execute user commands until Application Quit is requested.

static void cmd_processor( void )
{
  char					cmdbuf[MAXLEN];
  char					cmd[MAXLEN];
  char					arg[MAXLEN];
  char					*c;
  int					arg_val;

  // Display the command menu.
  show_menu();

  // Prompt for user command.
  prompt();

  // Repeat until Application Terminate is requested ...
  for ( ; ; )
  {
    // Input user command string, and if no errors encountered ...
    if ( fgets( cmdbuf, MAXLEN - 1, stdin ) != NULL )
    {
      // Truncate newline char at end of command string.
      cmdbuf[ strlen( cmdbuf ) - 1 ] = 0;

      // If command line is not blank ...
      if ( cmdbuf[0] != 0 )
      {
        // Force command string to lower case, except command=="text".
		if ( strncmp( cmdbuf, "text", 4 ) != 0 )
			for ( c = cmdbuf; *c != 0; c++ )
				*c = tolower( *c );

        // Parse command string into opcode and argument.
        cmd[0] = arg[0] = 0;
        sscanf( cmdbuf, "%s %s", cmd, arg );

        // Decode and execute command
        if ( exec_command( cmd, arg ) )
          break;
      }
      else
      {
        // Prompt for user command.
        prompt();
      }
    }
  }
}

///////////////////////////////////////////////////
// Set programmable attribute on a board.

static void set_attribute( struct BOARD *b, int val, int minval, int maxval, int ioctl_num )
{
  // Set attribute if value is in valid range, otherwise display error message.
  if ( ( val >= minval ) && ( val <= maxval ) )
    ioctl( b->handle, ioctl_num, val );
  else
  {
    printf( "Illegal or missing argument; must be in range %d..%d\n", minval, maxval );
    prompt();
  }
}


///////////////////////////////////////////////////////////////
// Set video system. We have to change some parameters for PAL.

static void set_vsystem( struct BOARD *b, int sys )
{
  // Set attribute if value is in valid range, otherwise display error message.
  if ( sys == SYS_PAL )
  {
    ioctl( b->handle, SX16_IOC_SET_VIDEO_SYSTEM, SYS_PAL );
    // optimized for PAL
    ioctl( b->handle, SX16_IOC_SET_M, 4 );
    ioctl( b->handle, SX16_IOC_SET_N, 1 );
  }
  else if ( sys == SYS_NTSC )
  {
    ioctl( b->handle, SX16_IOC_SET_VIDEO_SYSTEM, SYS_NTSC );
    // default settings
    ioctl( b->handle, SX16_IOC_SET_M, 15 );
    ioctl( b->handle, SX16_IOC_SET_N, 3 );
  }
  else
  {
    printf( "Illegal or missing argument; must be %d or %d\n", SYS_PAL, SYS_NTSC );
    prompt();
  }
}

/////////////////////////////////////////////////////
// Fetch and display board's current settings.

static char txt_muxformat	[][40] = { "svcd", "vcd", "dvd", "cdda", "video-es", "audio-es", "transport" };
static char txt_arate		[][40] = {	"", "32", "48", "56", "64", "80", "96", "112", "128", "160", "192", "224", "256", "320", "384" };
static char txt_mpeg		[][40] = { "mpeg2:D1-704", "mpeg2:2/3-D1", "mpeg2:1/2-D1", "mpeg1:SIF", "mpeg1:QSIF", "mpeg2:D1-720", "mpeg2:D1-640" };
static char txt_boolean		[][40] = { "off", "on" };
static char txt_vsystem		[][40] = { "", "pal", "ntsc" };
static char txt_intype		[][40] = { "s-video", "composite" };
static char txt_sysformat	[][40] = { "svcd", "vcd", "other" };
static char txt_avsel		[][40] = { "audio+video", "video only", "audio only" };
static char txt_asmprate	[][40] = { "44", "48", "32" };
static char txt_pcmres		[][40] = { "16", "18", "20", "24" };
static char txt_aformat		[][40] = { "mpeg1:layer-1", "mpeg1:layer-2" };
static char txt_cap_scale	[][40] = { "x1", "x2", "x2 (vertical only)", "x4" };
static char txt_cap_color	[][40] = { "white", "black", "red", "green", "blue",  "yellow", "cyan", "magenta" };
static char txt_cap_transp	[][40] = { "", "transparent", "black-background", "transp-text on white-bkgrnd", "black-text on white-bkgrnd" };
static char txt_cap_on		[][40] = { "off", "on" };

// Fetch and display the attributes.
static void get_attributes( struct BOARD *b )
{
	printf( "\nCURRENT SETTINGS -----------\n" );
	printf( "audio bass           = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_bass					) );
	printf( "audio treble         = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_treble				) );
	printf( "audio volume         = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_volume				) );
	printf( "video bitrate (bps)  = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_bitrate			) );
	printf( "video brightness     = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_brightness		) );
	printf( "video channel        = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_chan					) );
	printf( "video contrast       = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_contrast			) );
	printf( "gop                  = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_m						) );
	printf( "ip dist              = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_n						) );
	printf( "video saturation     = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_saturation		) );
	printf( "program number       = %d\n", ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_program_number			) );

	printf( "audio format         = %s\n", txt_aformat		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_format						) ] );
	printf( "a/v select           = %s\n", txt_avsel		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_av_select							) ] );
	printf( "audio samprate (KHz) = %s\n", txt_asmprate		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_sample_rate			) ] );
	printf( "audio pcm res (bits) = %s\n", txt_pcmres		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_pcm_resolution		) ] );
	printf( "video type           = %s\n", txt_intype		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_type							) ] );
	printf( "audio bitrate (kbps) = %s\n", txt_arate		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_bitrate					) ] );
	printf( "video colorbars      = %s\n", txt_boolean		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_colorbars_enable	) ] );
	printf( "video filter         = %s\n", txt_boolean		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_filter_enable		) ] );
	printf( "video system         = %s\n", txt_vsystem		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_system						) ] );
	printf( "variable bitrate     = %s\n", txt_boolean		[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_vbr_enable				) ] );
	printf( "system format        = %s\n", txt_sysformat	[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_system_format		) ] );
	printf( "mpeg type/picsize    = %s\n", txt_mpeg			[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_picsize					) ] );
	printf( "mux format           = %s\n", txt_muxformat	[ ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_mux_format							) ] );
}

// Fetch and display the caption settings.
static void get_caption_settings( struct BOARD *b )
{
	printf( "caption window       = %d\n", ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_window			) );
	printf( "caption X start      = %d\n", ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_xstart			) );
	printf( "caption X stop       = %d\n", ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_xstop				) );
	printf( "caption Y start      = %d\n", ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_ystart			) );
	printf( "caption Y stop       = %d\n", ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_ystop				) );
	printf( "caption scale        = %s\n", txt_cap_scale	[ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_scale	) ] );
	printf( "caption color        = %s\n", txt_cap_color	[ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_color	) ] );
	printf( "caption transparency = %s\n", txt_cap_transp	[ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_transp	) ] );
	printf( "caption on/off       = %s\n", txt_cap_on		[ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_on		) ] );
}

///////////////////////////////////////////////////
// Start MPEG data streaming to/from a file.

static void start_stream( struct BOARD *b, char *fname, int new_stream_mode )
{
  pthread_attr_t attr;

  // If we are already streaming ...
  if ( b->mode != MODE_IDLE )
  {
    // Abort command.
    printf( "already streaming\n" );
    prompt();
    return;
  }

  // Make the MPEG filename visible to the I/O thread.
  strcpy( b->fname, fname );

  // Indicate the new stream mode.
  b->mode = new_stream_mode;

  // Clear the stream thread's termination request.
  b->stop_rq = 0;

  // Create detached thread that will manage the MPEG stream flow.
  pthread_attr_init( &attr );
  pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
  pthread_create( &b->io_thread_id, NULL, &io_thread_func, b );
  pthread_attr_destroy( &attr );
}

//////////////////////////////////////////////////
// Halt MPEG data stream to/from a file.

static void stop_stream( struct BOARD *b )
{
  // If stream i/o thread is running ...
  if ( b->mode != MODE_IDLE )
  {
    // Request thread termination.
    b->stop_rq = 1;

    // Wait for the thread to terminate
    while ( b->mode != MODE_IDLE )
      sleep( 1 );
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////  MPEG STREAM I/O THREAD  //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

char mpeg_buf[MPEG_BUF_SIZE];


/////////////////////////////////////////////////////
// Copy a block of MPEG data from board to file.
// Returns number of bytes copied.

static int mpeg_to_file( struct BOARD *b, FILE *file )
{
  //char	mpeg_buf[MPEG_BUF_SIZE];
  int		len;

  //printf("%x %d\n", mpeg_buf, MPEG_BUF_SIZE);

  // Attempt to read MPEG data from board.
  len = read( b->handle, mpeg_buf, MPEG_BUF_SIZE );

  // Request Stream Halt and abort if we timed out while waiting to read MPEG data from board.
  if ( len <= 0 )
  {
    printf( "No signal\n" );
    prompt();
    b->stop_rq = 1;
    return 0;
  }

  // Write MPEG data to file.
  fwrite( mpeg_buf, 1, len, file );

  // Return number of bytes transferred.
  return len;
}

/////////////////////////////////////////////////////
// Copy a block of MPEG data from file to board.
// Returns number of bytes copied.

static int mpeg_from_file( struct BOARD *b, FILE *file )
{
  //char	mpeg_buf[MPEG_BUF_SIZE];
  int		len;
  int		remlen;
  int		nwritten;

  // Attempt to read MPEG data from file.
  len = fread( mpeg_buf, 1, MPEG_BUF_SIZE, file );

  // Request Stream Halt and abort if end of file reached.
  if ( len == 0 )
  {
    printf( "End of file reached\n" );
    prompt();
    b->stop_rq = 1;
    return 0;
  }

  // Repeat until all MPEG bytes have been transferred to the board ...
  for ( remlen = len; remlen > 0; )
  {
    // Attempt to write MPEG data to board.
    nwritten = write( b->handle, mpeg_buf, remlen );

    //  Request Stream Halt and abort if we timed out while writing MPEG data to board.
    if ( nwritten < 0 )
    {
      printf( "Player error\n" );
      prompt();
      b->stop_rq = 1;
      return 0;
    }

    // Compute remaining byte count.
    remlen -= nwritten;
  }

  // Return number of bytes transferred.
  return len;
}

////////////////////////////////////////////////////////////////////
// I/O THREAD MAIN.  Read/write MPEG data until stream is stopped.
// The arg is a pointer to the device's BOARD object.

void *io_thread_func( void *arg )
{
  FILE					*file;								// MPEG data file.
  unsigned int	rlen;									// MPEG bytes transferred.
  time_t				starttime;						// Stream start time.
  BOARD					*b = (BOARD *)arg;		// Cache pointer to board object.

  // Select the function that will be used to move MPEG data between file and board.
  int (* mpeg_transfer)( struct BOARD *b, FILE *file ) = ( b->mode == MODE_RECORD ) ? mpeg_to_file : mpeg_from_file;

  // Initialize stream statistics.
  b->mpeg_count = 0;		// No MPEG bytes have been transferred.
  time( &starttime );		// Log the start time.

  // Attempt to open the MPEG data file.  If successful ...
  if ( ( file = fopen( b->fname, ( b->mode == MODE_RECORD ) ? "wb" : "rb" ) ) != NULL )
  {
    // Report status.
    printf( "Starting stream\n");
    prompt();

    // Until stream termination is requested ...
    while ( !b->stop_rq )
    {
      // Transfer a block of MPEG data between file and board.  If this is the first
      // transfer, the driver will automatically start the stream running.
      rlen = mpeg_transfer( b, file );

      // Collect statistical info.
      b->mpeg_count += rlen;

      // Allow other processes to execute if we are in non-blocking mode and no data was available.
      if ( ( rlen == 0 ) && !BLOCKING_MODE )
      {
        // Sleep for approximately 10 milliseconds.
        struct timespec tv;
        tv.tv_sec = 0;
        tv.tv_nsec = 10000000;
        nanosleep( &tv, &tv );
      }
    }

    // Halt the stream.
    ioctl( b->handle, SX16_IOC_RW_STOP, 0 );

    // Close the MPEG data file.
    fclose( file );

    // Report statistics.
    printf( "%u MPEG bytes transferred\n", b->mpeg_count );
    prompt();
  }

  else
  {
    // Indicate there was a problem opening the MPEG data file.
    printf( "file open error\n" );
    prompt();
  }

  // Indicate stream is stopped.
  b->mode = MODE_IDLE;

  printf( "Stream stopped\n");
  prompt();

  // Terminate thread.
  return 0;
}


////////////////////////////////////////////////////////////////////////////////
// Video interrupt test. Wait for "num" interrupt, indicate time needed for it.

static void int_test( struct BOARD *b, int num )
{
    time_t t;
    int cnt;

    time(&t);
    for( cnt = 0; cnt < num; cnt++ )
        // This command waits for real interrupt or timeout if no interrupt detected.
        ioctl( b->handle, SX16_IOC_WAIT_VIDEO, 100 );   // 100 msec timeout.
    // Print result.
    printf( "%d seconds needed to process %d video interrupts.\n", time(0) - t, num );
    prompt();
}


