/* QBA module.
 * Copyright 2013 Sensoray Co. Inc.
 * Writing to file and modifying the stream
 * to comply with QBA (quad byte alignment) requirement
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <inttypes.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <time.h>
#include <sys/ioctl.h>
#include <assert.h>
#include "qba.h"

typedef u_int8_t u8;
typedef u_int16_t u16;
typedef int16_t s16;
typedef u_int32_t u32;
typedef int32_t s32;
typedef u_int64_t u64;
typedef int64_t s64;

#include "sx16ioctl.h"			    // ioctl definitions for sx16 driver.


static int fulld = 0;
static int copy_dump = 1;

// Video system.
#define SYS_PAL						1			// PAL.
#define SYS_NTSC					2			// NTSC.

// Video input type.
#define CVIDEO						0			// Composite video.
#define SVIDEO						1			// S-Video.

#define DISABLED 0
#define MAX_RBUF_SIZE (1024*300)  // will be much smaller

struct qba_state {
    int cur_pes; // current PES of interest
    int pes_len;
    int last_slice; // position of last slice
    int first_e0;
    int picture; // position of picture
    int search; // internal search marker
    int rbuf_pos; // buffer position
    int state;
    int last_pos; // position of last code
    int last_code;
    int skip_active;
    int cur_count; // current count
    int num_b8;
    int num_e0;
    int num_c0;


    int eof;
    u64 current_pos;
    u64 buffer_offset;
    u8 *buffer;
    u32 buffer_size;
    u64 overflow_offset;
    u8 *overflow_buffer;
    u32 overflow_size;
    u64 total_len;

    unsigned char *rbuf; // working buffer
    unsigned int rbuf_len; // working buffer len
    int oflow_err;
};


qba_handle qba_open(void)
{
    struct qba_state *qba;
    qba = (struct qba_state *) malloc(sizeof(struct qba_state));
    memset(qba, 0, sizeof(struct qba_state));
    qba->rbuf = (unsigned char *) malloc(MAX_RBUF_SIZE);
    if (qba->rbuf == NULL) {
        printf("out of memory error\n");
        return NULL;
    }
    return qba;
}

void qba_close(qba_handle qba)
{
    free(qba->rbuf);
    free(qba);
    return;
}

void qba_buf_put(qba_handle qba, unsigned char *buf, unsigned int size)
{
    if (qba->buffer)
        fprintf(stderr, "%s: warning: buffer already exists\n", __func__);
    qba->buffer = buf;
    qba->buffer_size = size;
    qba->buffer_offset = 0;// qba->current_pos;
    qba->eof = 0;
}

unsigned int qba_buf_get(qba_handle qba)
{
    u32 size;
    size = (u32)(qba->buffer_offset); //(current_pos - qba->buffer_offset);
    assert(size <= qba->buffer_size);
    qba->buffer = NULL;
    qba->buffer_size = 0;
    return size;
}



// returns position if found, otherwise 0
// state updated in qba
// searches for code starting at start
// len is size of buffer.
// code is code to search for
int find_start_code(unsigned char *buf, int start, int len, int code, struct qba_state *qba)
{
    int i = 0;
    int find_any = (code > 0xff) ? 1 : 0;
    
    for (i = start; i < len; i++) {
        switch (qba->search) {
        case 0:
        case 1:
            // search for 0x00
            if (buf[i] == 0x00) {
                qba->search++;
//                printf("[%d]found 0\n", i);
            } else {
                qba->search = 0;
            }
            break;
        case 2:
            // search for 0x01
            if (buf[i] == 0x01) {
                //printf("[%d]found 1\n", i);
                qba->search++;
            } else if (buf[i] != 0x00) {
                //printf("fail %x %x %x %x\n", buf[i-3], buf[i-2], buf[i-1], buf[i]);
                qba->search = 0;
            }
            break;
        case 3:
            if ((buf[i] == code) || find_any) {
                qba->search = 0;
                //printf("[%d]found 0x%x\n", i, buf[i]);
                return i-3;
            } else
                qba->search = 0;
            break;
        default:
            qba->search = 0;
            break;
        }
    }
    return 0;
}

// reverse find start code
int find_start_code_rev(unsigned char *buf, int start, int code)
{
    int i = 0;
    int state = 0;
    int find_any = (code > 0xff) ? 1 : 0;
    for (i = start-3; i >= 0; i--) {
        if ((buf[i] == 0x00) && (buf[i+1] == 0x00) && (buf[i+2] == 0x01) && (buf[i+3] == code))
            return i;
    }
    return -1;
}

void dump_bytes(unsigned char *buf, int bytes)
{
    int i;
    for (i = 0; i < bytes; i++) {
        printf(" 0x%02x", buf[i]);
        if (i % 8 == 7) {
            printf("\n");
        }
    }
    return;
}


int slicec = 0;

int flush_buffer(struct qba_state *qba)
{
    int copy_size;
    int flushed = 0;
#ifdef DEBUG
    if (qba->search) {
        printf("skipping %d bytes from flush\n");
    }
#endif

    copy_size = qba->rbuf_pos - qba->search;
    
    if ((qba->state != 0) || (copy_size < 0))
        return flushed;

    //printf("flush %d %d %d, size %d\n", qba->rbuf_pos, qba->state, qba->rbuf_len, copy_size);

    if (qba->buffer_offset + copy_size >= qba->buffer_size) {
        copy_size = qba->buffer_size - qba->buffer_offset;
        //printf("shorten copy size %d\n", copy_size);
    }

    memmove(qba->buffer + qba->buffer_offset, qba->rbuf, copy_size);
    //fwrite(mpeg_buf, 1, copy_size, file);
    memmove(qba->rbuf, qba->rbuf + copy_size, qba->rbuf_len - copy_size);
    
    qba->buffer_offset += copy_size;
    qba->rbuf_len -= copy_size;
    qba->rbuf_pos -= copy_size;
    qba->last_pos -= copy_size;
    qba->cur_pes -= copy_size;
    qba->last_slice -= copy_size;
    qba->last_pos -= copy_size;
    qba->first_e0 -= copy_size;
    qba->total_len +=  copy_size;
    return flushed;
}

int main(int argc, char **argv)
{
    qba_handle qba;
    int h; // handle to board
    size_t nr;
    char dev_name[100];
    unsigned char in[4096];
    unsigned char *out;
    int file_o;
    int hcopy;
    int nw;
    int total = 0;
    strcpy(dev_name, "/dev/sx16a0");
    // open board
    if (argc == 1) {
        printf("please enter file to save to as 2nd argument (\"qba file.mpg\")\n");
        return -EINVAL;
    }

    h = open( dev_name, O_RDWR);
    if (h == -1) {
        printf( "Driver open error\n" );
        return( -1 );
    }
    ioctl(h, SX16_IOC_SET_COLORBARS, DISABLED);
    ioctl(h, SX16_IOC_SET_VIDEO_SYSTEM, SYS_NTSC);
    // 2Mbps
    ioctl(h, SX16_IOC_SET_BITRATE, 2000000);
    // set to MPEG-2 D1 size
    ioctl(h, SX16_IOC_SET_MPEG,	0);
    ioctl(h, SX16_IOC_SET_INPUT, 0);
    ioctl(h, SX16_IOC_SET_INPUT_TYPE, CVIDEO);
    ioctl(h, SX16_IOC_SET_VBR, DISABLED);
    ioctl(h, SX16_IOC_SET_AUDIO_VOLUME, 6);
    ioctl(h, SX16_IOC_SET_AUDIO_MUTE, DISABLED);

#define OUTBUF_SIZE (16*1024)
    out = (unsigned char *) malloc(OUTBUF_SIZE);
    if (out == NULL)
        return -ENOMEM;


    file_o = open(argv[1], O_RDWR | O_CREAT, 0666);
    if (file_o == -1) {
        file_o = open(argv[1], O_RDWR | O_TRUNC, 0666);
    }
    if (file_o == -1) {
        printf("could not open file %s to write\n", argv[1]);
        free(out);
        close(h);
        return -1;
    }
    hcopy = -1;
    if (copy_dump) {
        hcopy = open("test_orig.mpg", O_RDWR | O_CREAT, 0666);
    }

    nr = read(h, in, 4096);
    if (hcopy != -1)
        write(hcopy, in, nr);
    qba = qba_open();
    if (qba == NULL) {
        free(out);
        close(h);
        close(file_o);
        return -ENOMEM;
    }
#define MAX_SIZE (4*1024*1024)
    printf("recording %d Mb to file.  Please wait\n", MAX_SIZE / 1024 / 1024);
    while (nr > 0 && total < MAX_SIZE) {
        qba_buf_put(qba, out, OUTBUF_SIZE);
        qba_write_mpegps(qba, in, nr);
        nw = qba_buf_get(qba);
        if (nw > 0) {
            //printf("writing to out file %d\n", nw);
            printf(".");
            fflush(stdin);
            write(file_o, out, nw);
            total += nw;
        }
        nr = read(h, in, 4096);
        if (hcopy != -1)
            write(hcopy, in, nr);
    }
    printf("\n");
    ioctl(h, SX16_IOC_RW_STOP, 0);
    close(h);
    if (hcopy != -1)
        close(hcopy);
    close(file_o);
    free(out);
    return 0;
}

static void process_qba(qba_handle qba);

// Init the system and execute user commands until termination.
void qba_write_mpegps(qba_handle qba, unsigned char *buf, unsigned int size)
{
  if (size < 0)
      return;
#ifdef DEBUG_MESSAGES
  printf("state: %d, rbuf_len %d, rbuf_pos %d, size %d\n",
         qba->state, qba->rbuf_len, qba->rbuf_pos, size);
#endif
  // QBA state. 

  if (qba->rbuf_len + size >= MAX_RBUF_SIZE) {
      printf("overflow!!\n");
      qba->oflow_err++;
      return;
  }
  // copy buffer to our working buffer
  memcpy(qba->rbuf + qba->rbuf_len, buf, size);
  qba->rbuf_len += size;

  while ((qba->buffer_offset < qba->buffer_size) && (qba->rbuf_pos < (qba->rbuf_len - 1024))) {
      process_qba(qba);
      if (qba->rbuf_len <= 100*1024)
          break;
  }
}

void process_qba(qba_handle qba)
{
    unsigned char *pbuf;
    int rc2;
    //printf("state %d pos:%d, len:%d\n", qba->state, qba->rbuf_pos, qba->rbuf_len);
    switch (qba->state) {
    case 0: 
        rc2 = find_start_code(qba->rbuf, qba->rbuf_pos, qba->rbuf_len, 0xe0, qba);
        assert (rc2 > qba->rbuf_pos);
        if (fulld)
            printf("state 0: rc2 %x\n", rc2);
        if (rc2 < 0) {
            // no PES found, increment rbuf_pos.
            flush_buffer(qba);
            printf("state 0 none found\n");
            return;
        } else {
            unsigned int nbytes;
            int left;
            int pes_len;
            if (qba->rbuf_pos >= qba->rbuf_len - 6) {
                flush_buffer(qba);
                return;
            }
            // found rc2
            qba->first_e0 = rc2;
            pbuf = qba->rbuf + rc2;
            pes_len = (pbuf[4] << 8) + pbuf[5];
            if (pbuf[8] == 0) {
                qba->rbuf_pos = rc2 + 7;
            }
            else {
                qba->rbuf_pos = rc2 + 9 + pbuf[8];
            }
            qba->cur_pes = rc2;
            qba->pes_len = pes_len;
            //printf("state 1 enter\n");
            qba->last_pos = rc2;
            qba->state = 1;
            //printf("rbuf pos:::: %d\n", qba->rbuf_pos);
        }
        break;
    case 1:
        // search for next picture start code from last_slice position.
        //printf("state 1: rbuf_len: %x \n", qba->rbuf_len);
        // 0xff + 1 == find any start code
        rc2 = find_start_code(qba->rbuf, qba->rbuf_pos, qba->rbuf_len, 0xff+1, qba);
        if (fulld)
            printf("state 1: %x\n", rc2);
        if (rc2 == 0) {
            return;
        }
        pbuf = qba->rbuf + rc2;

        switch (pbuf[3]) {
        case 0x1e:
            if (fulld)
                printf("to state 2\n");
            qba->state = 2;
            qba->last_slice = rc2;
            qba->cur_count = 0;
            qba->rbuf_pos = rc2 + 4;
            break;
        case 0xe0:
            //printf("how did we get here, slice %d !!\n", slicec);
            qba->rbuf_pos = rc2+4;
            qba->state = 0;
            flush_buffer(qba);
            break;
        case 0xba:
        case 0xc0:
        case 0xb3:
            // to state 0
            if (fulld)
                printf("to state 0\n");
            if (slicec == 471) {
                printf("\n");
            }
            qba->state = 0;
            qba->rbuf_pos = rc2+4;
            flush_buffer(qba);
            break;
        default:
            //
            qba->rbuf_pos = rc2 + 4;
            if (fulld)
                printf("default code: 0x%02x\n", pbuf[3]);
            flush_buffer(qba);
            break;
        }
        qba->last_pos = rc2;
        break;
    case 2:
        // found the slice code that we want.  now have to extend it.
        // search for the next start code.
        rc2 = find_start_code(qba->rbuf, qba->rbuf_pos, qba->rbuf_len, 0xff+1, qba);
        if (fulld)
            printf("state 2: %x\n", rc2);
        if (rc2 == 0) {
            return;
        }
        pbuf = qba->rbuf + rc2;
        if (fulld)
            printf("state 2: code %x\n", pbuf[3]);
        switch (pbuf[3]) {
        case 0x00:
        case 0xbe:
        case 0xb3:
        case 0xb5:
        case 0xb8:
            // found expected start code after last slice.  we 
            // can pad the last slice now.
            qba->state = 5;
            qba->rbuf_pos = rc2 + 4;
            qba->cur_count += (rc2 - qba->last_pos);
            if (fulld)
                printf("to state 5\n");
            break;
        case 0xba:
            // new pack code, go to state 3 (search for video packet)
            qba->state = 3;
            qba->rbuf_pos = rc2 + 4;
            qba->cur_count += (rc2 - qba->last_pos);
            if (fulld)
                printf("to state 3\n");
            break;
        case 0xe0:
        default:
            // found a new packet
            printf("unexpected packet\n");
            qba->state = 0;
            qba->rbuf_pos = rc2 + 4;
            flush_buffer(qba);
            break;
        }
        qba->last_pos = rc2;
        break;
    case 3: // search for e0 state
        rc2 = find_start_code(qba->rbuf, qba->rbuf_pos, qba->rbuf_len, 0xff+1, qba);
        if (fulld)
            printf("state 3 %x\n", rc2);
        if (rc2 == 0) {
            return;
        }
        pbuf = qba->rbuf + rc2;
        switch (pbuf[3]) {
        case 0xba:
        case 0xbb:
            qba->state = 3;
            qba->rbuf_pos = rc2 + 4;
            break;
        default:
            if (fulld)
                printf("to state 4 %x\n", pbuf[3]);
            qba->state = 4;
            qba->rbuf_pos = rc2 + 4;
            break;
        case 0xe0:
            if (fulld)
                printf("to state 2\n");
            qba->state = 2;
            qba->cur_pes = rc2;
            // skip the header
            qba->rbuf_pos = rc2 + 8;
            break;
        }
        qba->last_pos = rc2;
        break;
    case 4: // search for ba state
        rc2 = find_start_code(qba->rbuf, qba->rbuf_pos, qba->rbuf_len, 0xff+1, qba);
        if (rc2 == 0) {
            return;
        }
        pbuf = qba->rbuf + rc2;
        switch (pbuf[3]) {
        default:
            qba->state = 4;
            qba->rbuf_pos = rc2 + 4;
            break;
        case 0xe0:
            assert(0); // shouldn't get here
            break;
        case 0xba:
        case 0xbb:
            //printf("ba or bb found again\n");
            qba->state = 3;
            qba->rbuf_pos = rc2 + 4;
            break;
        }
        qba->last_pos = rc2;
        break;
    case 5:
    {
        // we are done
        unsigned char *pbuf;
        int offset;
        int pes_len;
        int diff = 9;
        int pes_headers = 0;
        // found rc2
        //qba->cur_count += rc2 - qba->last_pos;
        offset = qba->cur_count;
        rc2 = qba->last_pos;
        //printf("[%x][%d] slice slice is %x, %x,skip %x code %x %x %x %x\n", rc2, slicec, offset, offset % 4, 0, *(qba->rbuf+rc2+0),  *(qba->rbuf+rc2+1), *(qba->rbuf+rc2+2), *(qba->rbuf+rc2+3));

        // search backwards from here to find the associated PES header.
        // we need to know the PES length
        pbuf = qba->rbuf + qba->cur_pes;
        if (qba->cur_pes == qba->first_e0) {
            pes_headers += pbuf[8];
        } else {
            int other_pes;
            //dump_bytes(pbuf, slicec == 57 ? 32 : 16);
            pes_headers += 9 + pbuf[8];
            //printf("total PES headers %x\n", pes_headers);
            pbuf = qba->rbuf + qba->cur_pes;
              
        }

        offset -= pes_headers;
        offset = 4 - (offset % 4);
        // bytes to add
        offset = offset % 4;

        memmove(qba->rbuf + rc2 + offset, qba->rbuf + rc2, qba->rbuf_len - rc2);
        memset(qba->rbuf + rc2, 0, offset);

        // rbuf length increased
        qba->rbuf_len += offset; // rbuf length increased
        pes_len = (pbuf[4] << 8) + pbuf[5];
        //printf("pes length %d\n", pes_len);
        // increment the PES header by the inserted number of bytes.
        pes_len += offset;
        pbuf[4] = (pes_len >> 8) & 0xff;
        pbuf[5] = (pes_len & 0xff);
        qba->state = 0;
        qba->rbuf_pos = rc2 + offset;
        //printf("rbuf position %d\n", rc2);
        flush_buffer(qba);
        slicec++;
    }
    break;
    }
    return;
}

